/**
 * $Id: TraceGLE.java,v 1.1 1998/05/02 14:36:29 descarte Exp descarte $
 *
 * Copyright (c)1998 Arcane Technologies Ltd. <http://www.arcana.co.uk>
 *
 * OpenGL GLE Tubing/Extrusion method implementations for method tracing
 *
 * $Log: TraceGLE.java,v $
 * Revision 1.1  1998/05/02 14:36:29  descarte
 * Initial revision
 *
 *
 *      This software is Copyright (c)1998 Arcane Technologies Ltd.
 *      and is released under the ``Artistic'' licence which is
 *      available in the source distribution. If this license is not
 *      present, you have an unofficial release of this software.
 *      The official release may be downloaded from Arcane Technologies
 *      Ltd. WWW site at:
 *
 *              http://www.arcana.co.uk/products/shapeshifter
 *
 */

package com.hermetica.gle;

import com.hermetica.util3d.Streams;

/**
 * Implementation of the routines exposed by the GLE library that simply
 * trace execution through the various GLE calls. A CoreGLE object should be
 * attached to instances of this class to ensure actual execution of the
 * methods occurs.
 * <P>
 * @version $Id: TraceGLE.java,v 1.1 1998/05/02 14:36:29 descarte Exp descarte $
 * @author Alligator Descartes &lt;<A HREF="http://www.arcana.co.uk">http://www.arcana.co.uk</A>&gt;
 */
public class TraceGLE implements GLE {

    /** Version information */
    public static final String VERSION = 
        new String( "$Id: TraceGLE.java,v 1.1 1998/05/02 14:36:29 descarte Exp descarte $" );

    /** The parent object through which calls are routed */
    private GLE parent = null;

    /** 
     * Default pipeline constructor where the pipeline will have its
     * parent set at a later date with setParent()
     */
    public TraceGLE() {
      }

    /**
     * Constructs a new TraceGLE pipeline with the given pipeline object as
     * a parent in the execution stack.
     * <P>
     * @param parent The parent pipeline object
     */
    public TraceGLE( GLE parent ) {
        setParent( parent );
      }

    /**
     * Sets the parent interface.
     * <P>
     * @param parent The parent interface
     */
    public void setParent( GLE parent ) {
        this.parent = parent;
      }

    /** 
     * Returns the parent interface
     */
    public GLE getParent() {
        return this.parent;
      }

    /** The operating mode */
    private int mode = SUMMARY;

    /** 
     * Sets the operating mode of this pipeline object. The default mode
     * is TraceGLE.SUMMARY
     * <P>
     * @param mode Either TraceGLE.SUMMARY or TraceGLE.VERBOSE
     */
    public void setMode( int mode ) {
        this.mode = mode;
      }

    /**
     * Returns the current operating mode of the current pipeline object
     * <P>
     */
    public int getMode() {
        return this.mode;
      }

    /**
     * Dumps the first 8 ( or maxiumum ) elements of an array
     * <P>
     * @param ary The array to dump information for
     */
    private String dumpArray( float[] ary ) {
        StringBuffer sb = new StringBuffer();
        if ( ary != null ) {
            int max = ( ary.length <= 8 ) ? ary.length : 8;
            sb.append( "[ " );
            for ( int i = 0 ; i < max ; i++ ) {
                sb.append( ary[i] );
                sb.append( ", " );
              }
            sb.append( "... ]" );
          }
        return sb.toString();
      }

    /**
     * Dumps the first 8 ( or maxiumum ) elements of an array
     * <P>
     * @param ary The array to dump information for
     */
    private String dumpArray( double[] ary ) {
        StringBuffer sb = new StringBuffer();
        if ( ary != null ) {
            int max = ( ary.length <= 8 ) ? ary.length : 8;
            sb.append( "[ " );
            for ( int i = 0 ; i < max ; i++ ) {
                sb.append( ary[i] );
                sb.append( ", " );
              }
            sb.append( "... ]" );
          }
        return sb.toString();
      }

    /**
     * Dumps the first 8 ( or maxiumum ) elements of an array
     * <P>
     * @param ary The array to dump information for
     */
    private String dumpArray( float[][] ary ) {
        return dumpArray( ary[0] );
      }

    /**
     * Dumps the first 8 ( or maxiumum ) elements of an array
     * <P>
     * @param ary The array to dump information for
     */
    private String dumpArray( double[][] ary ) {
        return dumpArray( ary[0] );
      }

    /**
     * Dumps the first 8 ( or maxiumum ) elements of an array
     * <P>
     * @param ary The array to dump information for
     */
    private String dumpArray( float[][][] ary ) {
        return dumpArray( ary[0][0] );
      }

    /**
     * Dumps the first 8 ( or maxiumum ) elements of an array
     * <P>
     * @param ary The array to dump information for
     */
    private String dumpArray( double[][][] ary ) {
        return dumpArray( ary[0][0] );
      }

    /**
     * Returns the current join style for connected extruded segments
     * <P>
     */
    public final int gleGetJoinStyle() {
        int rv = 0;
        Streams.writeTrace( "gleGetJoinStyle()" );
        if ( parent != null ) {
            rv = parent.gleGetJoinStyle();
          }
        return rv;
      }

    /**
     * Sets the current join style for connected extruded segments.
     * <P>
     * @param style The new join style 
     */
    public final void gleSetJoinStyle( int style ) {
        if ( getMode() == SUMMARY ) {
            Streams.writeTrace( "gleSetJoinStyle()" );
          } else {
            Streams.writeTrace( "gleSetJoinStyle( " + style + " )" );
          }
        if ( parent != null ) {
            parent.gleSetJoinStyle( style );
          }
      }

    /**
     * Sets the current texture coordinate generation mode
     * <P>
     * @param mode Bitwise OR of texture flags
     */
    public final void gleTextureMode( int mode ) {
        if ( getMode() == SUMMARY ) {
            Streams.writeTrace( "gleTextureMode()" );
          } else {
            Streams.writeTrace( "gleTextureMode( " + mode + " )" );
          }
        if ( parent != null ) {
            parent.gleTextureMode( mode );
          }
      }

    /**
     * Draws a polycylinder, specified as a polyline
     * <P>
     * @param npoints The number of points in the polyline
     * @param pointArray An array of points specified as ( x, y, z )
     * @param colourArray An array of colors at each polyline vertex specified as RGB
     * @param radius The radius of the polycylinder
     */
    public void glePolyCylinder( int npoints, 
                                 double[][] pointArray,
                                 float[][] colourArray, 
                                 double radius ) throws GLEException {

        if ( getMode() == SUMMARY ) {
            Streams.writeTrace( "glePolyCylinder()" );
          } else {
            Streams.writeTrace( "glePolyCylinder( " + npoints + ", " + dumpArray( pointArray ) +
                                ", " + dumpArray( colourArray ) + ", " + radius + " )" );
          }
        if ( parent != null ) {
            parent.glePolyCylinder( npoints, pointArray, colourArray, radius );
          }
      }

    /** 
     * Draws a polycone specified as a polyline with radii
     * <P>
     * @param npoints The number of points in the polyline
     * @param pointArray An array of points specified as ( x, y, z )
     * @param colourArray An array of colours at each polyline vertex specified as RGB
     * @param radiusArray An array containing the radii of the cone at each polyline vertex
     */
    public void glePolyCone( int npoints,
                             double[][] pointArray,
                             float[][] colourArray,
                             double[] radiusArray ) throws GLEException {

        if ( getMode() == SUMMARY ) {
            Streams.writeTrace( "glePolyCone()" );
          } else {
            Streams.writeTrace( "glePolyCone( " + npoints + ", " + dumpArray( pointArray ) +
                                ", " + dumpArray( colourArray ) + ", " +
                                dumpArray( radiusArray ) + " )" );
          }
        if ( parent != null ) {
            parent.glePolyCone( npoints, pointArray, colourArray, radiusArray );
          }
      }

    /**
     * Extrudes an arbitrary 2D contour along an arbitrary 3D path.
     * <P>
     * @param ncp The number of contour points
     * @param contour An array containing the points forming the 2D contour specified as ( x, y )
     * @param contourNormal An array containing the normals for each contour point
     * @param up The up vector for the contour
     * @param npoints The number of points in the 3D polyline
     * @param pointArray An array containing the vertices for the 3D polyline specified as ( x, y, z )
     * @param colourArray An array containing the colours at each polyline vertex specified as RGB
     */
    public void gleExtrusion( int ncp,
                              double[][] contour,
                              double[][] contourNormal,
                              double[] up,
                              int npoints,
                              double[][] pointArray,
                              float[][] colourArray ) throws GLEException {

        if ( getMode() == SUMMARY ) {
            Streams.writeTrace( "gleExtrusion()" );
          } else {
            Streams.writeTrace( "gleExtrusion( " + ncp + ", " + dumpArray( contour ) + ", " +
                                dumpArray( contourNormal ) + ", " +
                                dumpArray( up ) + ", " + npoints + ", " +
                                dumpArray( pointArray ) + ", " +
                                dumpArray( colourArray ) + " )" );
          }
        if ( parent != null ) {
            parent.gleExtrusion( ncp, contour, contourNormal, up, npoints,
                                 pointArray, colourArray );
          }
      }

    /**
     * Extrudes an arbitrary 2D contour along a 3D path specifying local
     * rotations ( twists ) as the contour is extruded.
     * <P>
     * @param ncp The number of contour points
     * @param contour An array containing the points forming the 2D contour specified as ( x, y )
     * @param contourNormal An array containing the normals for each contour point
     * @param up The up vector for the contour
     * @param npoints The number of points in the 3D polyline
     * @param pointArray An array containing the vertices for the 3D polyline specified as ( x, y, z )
     * @param colourArray An array containing the colours at each polyline vertex specified as RGB
     * @param twistArray An array containing the twists to be applied at each contour point ( in degrees )
     */
    public void gleTwistExtrusion( int ncp,
                                   double[][] contour,
                                   double[][] contourNormal,
                                   double[] up,
                                   int npoints,
                                   double[][] pointArray,
                                   float[][] colourArray,
                                   double[] twistArray ) throws GLEException {

        if ( getMode() == SUMMARY ) {
            Streams.writeTrace( "gleTwistExtrusion()" );
          } else {
            Streams.writeTrace( "gleTwistExtrusion( " + ncp + ", " + dumpArray( contour ) + ", " +
                                dumpArray( contourNormal ) + ", " +
                                dumpArray( up ) + ", " + npoints + ", " +
                                dumpArray( pointArray ) + ", " +
                                dumpArray( colourArray ) + ", " +
                                dumpArray( twistArray ) + " )" );
          }
        if ( parent != null ) {
            parent.gleTwistExtrusion( ncp, contour, contourNormal, up, npoints,
                                      pointArray, colourArray, twistArray );
          }
      }

    /**
     * Extrudes an arbitrary 2D contour along a 3D path specifying local
     * affine transformations as the contour is extruded.
     * <P>
     * @param ncp The number of contour points
     * @param contour An array containing the points forming the 2D contour specified as ( x, y )
     * @param contourNormal An array containing the normals for each contour point
     * @param up The up vector for the contour
     * @param npoints The number of points in the 3D polyline
     * @param pointArray An array containing the vertices for the 3D polyline specified as ( x, y, z )
     * @param colourArray An array containing the colours at each polyline vertex specified as RGB
     * @param xformArray An array containing the affine transformations to be applied at each contour point ( in degrees )
     */
    public void gleSuperExtrusion( int ncp,
                                   double[][] contour,
                                   double[][] contourNormal,
                                   double[] up,
                                   int npoints,
                                   double[][] pointArray,
                                   float[][] colourArray,
                                   double[][][] xformArray ) throws GLEException {

        if ( getMode() == SUMMARY ) {
            Streams.writeTrace( "gleSuperExtrusion()" );
          } else {
            Streams.writeTrace( "gleSuperExtrusion( " + ncp + ", " + dumpArray( contour ) + ", " +
                                dumpArray( contourNormal ) + ", " +
                                dumpArray( up ) + ", " + npoints + ", " +
                                dumpArray( pointArray ) + ", " +
                                dumpArray( colourArray ) + ", " +
                                dumpArray( xformArray ) + " )" );
          }
        if ( parent != null ) {
            parent.gleSuperExtrusion( ncp, contour, contourNormal, up, npoints,
                                      pointArray, colourArray, xformArray );
          }
      }

    /**
     * Sweep an arbitrary contour along a helical path.
     * <P>
     * @param ncp The number of contour points
     * @param contour An array containing the points forming the 2D contour specified as ( x, y )
     * @param contourNormal An array containing the normals for each contour point
     * @param up The up vector for the contour
     * @param startRadius The spiral starts in XY plane
     * @param drdTheta The change in radius per revolution
     * @param startZ The starting Z value
     * @param dzdTheta The change in Z per revolution
     * @param startTransform The starting contour affine transformation
     * @param dTransformdTheta The tangent change transformation per revolution
     * @param startTheta The start angle in XY plane
     * @param sweepTheta The number of degrees to spiral around
     */
    public void gleSpiral( int ncp,
                           double[][] contour,
                           double[][] contourNormal,
                           double[] up,
                           double startRadius,
                           double drdTheta,
                           double startZ,
                           double dzdTheta,
                           double[][] startTransform,
                           double[][] dTransformdTheta,
                           double startTheta,
                           double sweepTheta ) throws GLEException {

        if ( getMode() == SUMMARY ) {
            Streams.writeTrace( "gleSpiral()" );
          } else {
            Streams.writeTrace( "gleSpiral( " + ncp + ", " + dumpArray( contour ) + ", " +
                                dumpArray( contourNormal ) + ", " + dumpArray( up ) +
                                ", " + startRadius + ", " + drdTheta + ", " +
                                startZ + ", " + dzdTheta + ", " + dumpArray( startTransform ) +
                                ", " + dumpArray( dTransformdTheta ) + ", " + startTheta +
                                ", " + sweepTheta + " )" );
          }
        if ( parent != null ) {
            parent.gleSpiral( ncp, contour, contourNormal, up, startRadius,
                              drdTheta, startZ, dzdTheta, startTransform,
                              dTransformdTheta, startTheta, sweepTheta );
          }
      }

    /**
     * Sweep an arbitrary contour along a helical path. The sweep will be
     * performed as a shear along the z-axis so that the orientation of the
     * contour is displaced, rather than translated, as the contour is swept.
     * <P>
     * @param ncp The number of contour points
     * @param contour An array containing the points forming the 2D contour specified as ( x, y )
     * @param contourNormal An array containing the normals for each contour point
     * @param up The up vector for the contour
     * @param startRadius The spiral starts in XY plane
     * @param drdTheta The change in radius per revolution
     * @param startZ The starting Z value
     * @param dzdTheta The change in Z per revolution
     * @param startTransform The starting contour affine transformation
     * @param dTransformdTheta The tangent change transformation per revolution
     * @param startTheta The start angle in XY plane
     * @param sweepTheta The number of degrees to spiral around
     */
    public void gleLathe( int ncp,
                          double[][] contour,
                          double[][] contourNormal,
                          double[] up,
                          double startRadius,
                          double drdTheta,
                          double startZ,
                          double dzdTheta,
                          double[][] startTransform,
                          double[][] dTransformdTheta,
                          double startTheta,
                          double sweepTheta ) throws GLEException {

        if ( getMode() == SUMMARY ) {
            Streams.writeTrace( "gleLathe()" );
          } else {
            Streams.writeTrace( "gleLathe( " + ncp + ", " + dumpArray( contour ) + ", " +
                                dumpArray( contourNormal ) + ", " + dumpArray( up ) +
                                ", " + startRadius + ", " + drdTheta + ", " +
                                startZ + ", " + dzdTheta + ", " + dumpArray( startTransform ) +
                                ", " + dumpArray( dTransformdTheta ) + ", " + startTheta +
                                ", " + sweepTheta + " )" );
          }
        if ( parent != null ) {
            parent.gleLathe( ncp, contour, contourNormal, up, startRadius,
                             drdTheta, startZ, dzdTheta, startTransform,
                             dTransformdTheta, startTheta, sweepTheta );
          }
      }

    /**
     * Generalized Torus. Similar to gleSpiral() except contour is a circle.
     * Uses gleSpiral() to draw.
     * <P>
     * @param rToroid The circle contour radius
     * @param startRadius The spiral starts in XY plane
     * @param drdTheta The change in radius per revolution
     * @param startZ The starting Z value
     * @param dzdTheta The change in Z per revolution
     * @param startTransform The starting contour affine transformation
     * @param dTransformdTheta The tangent change transformation per revolution
     * @param startTheta The start angle in XY plane
     * @param sweepTheta The number of degrees to spiral around
     * @see com.hermetica.gle.GLE#gleSpiral
     */
    public void gleHelicoid( double rToroid,
                             double startRadius,
                             double drdTheta,
                             double startZ,
                             double dzdTheta,
                             double[][] startTransform,
                             double[][] dTransformdTheta,
                             double startTheta,
                             double sweepTheta ) throws GLEException {

        if ( getMode() == SUMMARY ) {
            Streams.writeTrace( "gleHelicoid()" );
          } else {
            Streams.writeTrace( "gleHelicoid( " + rToroid + ", " + startRadius + ", " + 
                                drdTheta + ", " + startZ + ", " + dzdTheta + ", " +
                                dumpArray( startTransform ) + ", " + 
                                dumpArray( dTransformdTheta ) + ", " + startTheta +
                                ", " + sweepTheta + " )" );
          }
        if ( parent != null ) {
            parent.gleHelicoid( rToroid, startRadius, drdTheta, startZ, dzdTheta, 
                                startTransform, dTransformdTheta, startTheta,
                                sweepTheta );
          }
      }

    /**
     * Generalized Torus. Similar to gleLathe() except contour is a circle.
     * Uses gleLathe() to draw.
     * <P>
     * @param rToroid The circle contour radius
     * @param startRadius The spiral starts in XY plane
     * @param drdTheta The change in radius per revolution
     * @param startZ The starting Z value
     * @param dzdTheta The change in Z per revolution
     * @param startTransform The starting contour affine transformation
     * @param dTransformdTheta The tangent change transformation per revolution
     * @param startTheta The start angle in XY plane
     * @param sweepTheta The number of degrees to spiral around
     * @see com.hermetica.gle.GLE#gleSpiral
     */
    public void gleToroid( double rToroid,
                           double startRadius,
                           double drdTheta,
                           double startZ,
                           double dzdTheta,
                           double[][] startTransform,
                           double[][] dTransformdTheta,
                           double startTheta,
                           double sweepTheta ) throws GLEException {

        if ( getMode() == SUMMARY ) {
            Streams.writeTrace( "gleToroid()" );
          } else {
            Streams.writeTrace( "gleToroid( " + rToroid + ", " + startRadius + ", " + 
                                drdTheta + ", " + startZ + ", " + dzdTheta + ", " +
                                dumpArray( startTransform ) + ", " + 
                                dumpArray( dTransformdTheta ) + ", " + startTheta +
                                ", " + sweepTheta + " )" );
          }
        if ( parent != null ) {
            parent.gleToroid( rToroid, startRadius, drdTheta, startZ, dzdTheta, 
                              startTransform, dTransformdTheta, startTheta,
                              sweepTheta );
          }
      }

    /**
     * Draws screw-type shapes. Takes a contour and extrudes it along the
     * z-axis from a start z-value of <B>startz</B> to and end z-value
     * of <B>endz</B>. During the extrusion, it will spin the contour along
     * the contour origin by <B>twist</B> degrees.
     * <P>
     * @param ncp The number of contour points
     * @param contour An array containing the vertex data for the 2D contour specified as ( x, y )
     * @param contourNormal An array containing the normal data for the 2D contour
     * @param up The up vector for the contour
     * @param startz The start z-value for the segment
     * @param endz The end z-value for the segment
     * @param twist The number of rotations to be applied to the contour
     */
    public void gleScrew( int ncp,
                          double[][] contour,
                          double[][] contourNormal,
                          double[] up,
                          double startz,
                          double endz,
                          double twist ) throws GLEException {

        if ( getMode() == SUMMARY ) {
            Streams.writeTrace( "gleScrew()" );
          } else {
            Streams.writeTrace( "gleScrew( " + ncp + ", " + dumpArray( contour ) + ", " +
                                dumpArray( contourNormal ) + ", " +
                                dumpArray( up ) + ", " + startz + ", " + endz + ", " +
                                twist + " )" );
          }
        if ( parent != null ) {
            parent.gleScrew( ncp, contour, contourNormal, up, startz, endz,
                             twist );
          }
      }
  }
